/**
 *    @file  libhttp.c
 *   @brief  HTTP parsing and building library
 *
 *  @author  François Cerbelle (Fanfan), francois@cerbelle.net
 *
 *  @internal
 *       Created:  15/11/2024
 *      Revision:  none
 * Last modified:  2024-11-22 08:26
 *      Compiler:  gcc
 *  Organization:  Cerbelle.net
 *     Copyright:  Copyright (c) 2024, François Cerbelle
 *
 *  This source code is released for free distribution under the terms of the
 *  GNU General Public License as published by the Free Software Foundation.
 */

#include "libhttp.h"
#include "sclist.h"
#include "base64.h"                             /* Base64 encoder and decoder */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

typedef struct HTTP_s {
    sclist_t* headers;
    char* body;
} HTTP_t;

typedef struct HTTPHeader_s {
    sclistrecord_t* sclistrecord;
    char* name;
    char* value;
} HTTPHeader_t;

HTTPHeader_t* HTTPHeader_new(const char* name, const char* value) {
    HTTPHeader_t* header;

    assert(name);
    assert(value);

    if (NULL==(header=malloc(sizeof(struct HTTPHeader_s)))) {
        perror("HTTPHeader_new");
        return NULL;
    }
    if (NULL==(header->name = strdup(name))) {
        perror("HTTPHeader_new name");
        free(header);
        return NULL;
    }
    if (NULL==(header->value = strdup(value))) {
        perror("HTTPHeader_new value");
        free(header->name);
        free(header);
        return NULL;
    }
    header->sclistrecord = NULL;
    return header;
}

void          HTTPHeader_del(HTTPHeader_t* header) {
    assert(header);
    assert(header->name);
    assert(header->value);

    free(header->name);
    free(header->value);
    free(header);
}

HTTPHeader_t* HTTPHeader_basicauth(const char* login, const char* pass) {
    char* auth_encoded;
    char* auth;

    assert(login);
    assert(pass);

    if (NULL==(auth=malloc(strlen("Basic ")+strlen(login)+1+strlen(pass)+1))) {
        perror("HTTPHeader_basicauth");
        return NULL;
    }
    /* strcpy/strcat expected to be faster than sprintf */
    strcpy(auth,login);
    strcat(auth,":");
    strcat(auth,pass);
    auth_encoded=base64_encode(auth);
    free(auth);

    if (NULL==(auth=malloc(strlen("Basic ")+strlen(auth_encoded)+1))) {
        perror("HTTPHeader_basicauth");
        free(auth_encoded);
        return NULL;
    }
    strcpy(auth,"Basic ");
    strcat(auth, auth_encoded);

    return HTTPHeader_new("Authorization",auth);
}

char*         HTTPHeader_getname(HTTPHeader_t* header) {
    assert(header);
    return header->name;
}

char*         HTTPHeader_getvalue(HTTPHeader_t* header) {
    assert(header);
    return header->value;
}

HTTP_t* HTTP_new() {
    HTTP_t* retval;

    if (NULL==(retval=malloc(sizeof(struct HTTP_s)))) {
        perror("HTTP_new HTTP_s");
        return NULL;
    }

    if (NULL==(retval->body=malloc(1))) {
        perror("HTTP_new body");
        free(retval);
        return NULL;
    };
    retval->body[0]=0;

    if (NULL==(retval->headers = sclist_new())) {
        perror("HTTP_new headers");
        free(retval->body);
        free(retval);
        return NULL;
    }

    return retval;
}

void HTTP_del(HTTP_t* http) {
    sclistrecord_t* current;

    assert(http);
    assert(http->headers);
    assert(http->body);

    current = sclist_firstrecord(http->headers);
    while (current) {
        HTTPHeader_t* header=sclist_getvalue(current);
        if (header->name)
            free(header->name);
        if (header->value)
            free(header->value);
        sclistrecord_t* tmp = current;
        current = sclist_nextrecord(current);
        sclist_remrecord(http->headers,tmp);
    }

    free(http->body);
    free(http);
}

char* HTTP_setbody(HTTP_t* http, const char* body) {
    char* newbody;

    assert(http);
    assert(body);
    assert(http->body);

    if (NULL==(newbody=realloc(http->body,strlen(body)+1))) {
        perror("HTTP_setbody");
        return NULL;
    }

    http->body = newbody;
    strcpy(http->body, body);
    return http->body;
}

char* HTTP_getbody(HTTP_t* http) {
    assert(http);
    assert(http->body);

    return http->body;
}

HTTP_t* HTTP_addheader(HTTP_t* http, HTTPHeader_t* header) {
    assert(http);
    assert(http->headers);
    assert(http->body);
    assert(header);

    /* Header already belonging to an header list */
    if (header->sclistrecord) {
        perror("HTTP_addheader, header already in a list");
        return NULL;
    }

    if (NULL==(header->sclistrecord=sclist_addrecord(http->headers,header))) {
        perror("HTTP_addheader");
        return NULL;
    }
    return http;
}

HTTPHeader_t* HTTP_firstheader(const HTTP_t* http) {
    assert(http);
    assert(http->headers);
    sclistrecord_t* header_item=sclist_firstrecord(http->headers);
    if (!header_item)
        return NULL;
    return sclist_getvalue(header_item);
}

HTTPHeader_t* HTTP_nextheader(const HTTPHeader_t* header) {
    sclistrecord_t* list_it;

    assert(header);

    list_it = header->sclistrecord;

    /* EOL reached */
    if (NULL==(list_it = sclist_nextrecord(list_it)))
        return NULL;
    else
        return sclist_getvalue(list_it);
}

HTTPHeader_t* HTTP_findheader(const HTTPHeader_t* start, const char* name, const char* value) {
    sclistrecord_t* header_it;

    assert(start);
    assert(start->sclistrecord);
    assert(name);
    /* value can be null for "any" */

    header_it = start->sclistrecord;

    /* Find the record pointer in the list */
    for (; header_it; header_it=sclist_nextrecord(header_it)) {
        HTTPHeader_t* header_ptr = sclist_getvalue(header_it);
        assert(header_ptr);
        char* header_name = HTTPHeader_getname(header_ptr);
        if (0==strcmp(header_name,name)) {
            if (NULL==value)
                break;
            char* header_value = HTTPHeader_getvalue(header_ptr);
            assert(header_value);
            if (0==strcmp(header_value,value))
                break;
        }
    }
    if (!header_it)
        return NULL;
    else
        return sclist_getvalue(header_it);
}

HTTP_t* HTTP_delheader(HTTP_t* http, HTTPHeader_t* header) {
    assert(http);
    assert(http->headers);
    assert(http->body);
    assert(header);

    /* Calling function has the header pointer, can and need to free it */
    sclist_remrecord(http->headers, header->sclistrecord);
    /** @todo: make sclist_remrecord return a status and use it */
#if 0
    if (NULL==sclist_remrecord(http->headers, header->sclistrecord)) {
        perror("HTTP_addheader");
        return NULL;
    }
#endif
    header->sclistrecord = NULL;

    return http;
}

char* HTTP_buildrequest(const HTTPMethod_t method, const char* uri, const HTTPVersion_t version) {
    char* retval;

    assert(method<=HTTPMETHOD_INVALID);
    assert(version<=HTTPVERSION_INVALID);

    /* NULL URI allowed, as an empty string */
    if (NULL==(retval=malloc(7+1+(uri?strlen(uri)+1:0)+8+1))) {
        perror("HTTP_getrequest");
        exit(EXIT_FAILURE);
    }
    retval[0]=0;

    switch(method) {
    case HTTPMETHOD_GET:
        strcat(retval, "GET");
        break;;
    case HTTPMETHOD_HEAD:
    case HTTPMETHOD_POST:
    case HTTPMETHOD_PUT:
    case HTTPMETHOD_DELETE:
    case HTTPMETHOD_CONNECT:
    case HTTPMETHOD_OPTIONS:
    case HTTPMETHOD_TRACE:
    case HTTPMETHOD_PATCH:
    case HTTPMETHOD_INVALID:
    default:
        fprintf(stderr,"HTTP method not supported\n");
        exit(EXIT_FAILURE);
        break;;
    }

    /* NULL URI allowed, as an empty string */
    if (uri) {
        strcat(retval," ");
        strcat(retval, uri);
    }

    strcat(retval," HTTP/");
    switch(version) {
    case HTTPVERSION_HTTP11:
    case HTTPVERSION_HTTP11b:
        strcat(retval, "1.1");
        break;;
    case HTTPVERSION_HTTP09:
    case HTTPVERSION_HTTP10:
    case HTTPVERSION_HTTP2:
    case HTTPVERSION_HTTP3:
    case HTTPVERSION_INVALID:
    default:
        fprintf(stderr,"HTTP version not supported\n");
        exit(EXIT_FAILURE);
        break;;
    }
    strcat(retval, "\r\n");
    return retval;
}

static char* strconcat(char* dst, char* src) {
    char* newdst;

    assert(dst);
    assert(src);

    if (NULL==(newdst=realloc(dst,strlen(dst)+strlen(src)+1))) {
        perror("strconcat");
        return NULL;
    }
    dst = newdst;
    strcat(dst,src);
    return dst;
}

char*         HTTP_buildheaders(const HTTP_t* http) {
    char* headers=NULL;
    sclistrecord_t* sclistrecord;
    HTTPHeader_t* header;

    assert(http);
    assert(http->headers);
    assert(http->body);

    if (NULL==(headers=strdup(""))) {
        perror("HTTP_buildheaders headers");
        return NULL;
    }
    for (sclistrecord=sclist_firstrecord(http->headers); sclistrecord; sclistrecord = sclist_nextrecord(sclistrecord)) {
        header = sclist_getvalue(sclistrecord);
        assert(header);
        /** @todo: OOM tests */
        headers=strconcat(headers,HTTPHeader_getname(header));
        headers=strconcat(headers,": ");
        headers=strconcat(headers,HTTPHeader_getvalue(header));
        headers=strconcat(headers,"\r\n");
    }


    return headers;
}
